import inspect
from functools import wraps
from inspect import cleandoc
from typing import Callable, Literal

from .utils import rscompile, load_elf, populate_from_elf, elf_add_trampoline, asm_preamble

from .core import Raspsim


def _asm_or_c_decorator(
    fn: Callable[[], Raspsim],
    lang: Literal["c", "assembler"],
    code_formatter: Callable[[str], str] | None = None,
) -> Callable[[], Raspsim]:
    if not callable(fn):
        raise TypeError("Argument must be a function or a string")

    spec = inspect.getfullargspec(fn)
    if spec.args:
        raise TypeError("Function must not have arguments")

    ret = spec.annotations.get("return")
    if ret is None or not issubclass(ret, Raspsim):
        raise TypeError("Function must return a Raspsim instance")

    name = fn.__name__
    code = fn.__doc__
    if code is None:
        raise ValueError("Function must have a docstring")

    if code_formatter is not None:
        code = code_formatter(code)

    @wraps(fn)
    def wrapper() -> Raspsim:
        with rscompile(code, entry_label=name, lang=lang) as f:
            elf = load_elf(f)

        elf_add_trampoline(elf)
        sim = Raspsim()
        populate_from_elf(sim, elf)
        sim.run()
        return sim

    return wrapper


def asm(fn: Callable[[], Raspsim]) -> Callable[[], Raspsim]:
    """
    A decorator that formats a function's code as Intel syntax assembly and
    applies an assembler-specific decorator.
    Args:
        fn (Callable[[], Raspsim]): The function to be decorated.
    Returns:
        Callable[[], Raspsim]: The decorated function with assembly formatting.

    Example:
        @asm
        def one() -> Raspsim:
            \"\"\"
            mov rax, 1
            ret
            \"\"\"

        print(one().rax)
    """

    def formatter(code: str):
        return asm_preamble(fn.__name__) + code + "\n"

    return _asm_or_c_decorator(fn, "assembler", formatter)



import subprocess
import tempfile
from typing import Literal, IO, Generator
from dataclasses import dataclass, field
from contextlib import contextmanager
import logging
from inspect import cleandoc

from elftools.elf.elffile import ELFFile

from .core import RaspsimException, Prot, getProtFromELFSegment, Raspsim

# "-g",
# "-gdwarf-4",
# "-O3",
# "-Wall",
# "-Wextra",
# "-Wpedantic",
# "-pedantic-errors",

# if segment["p_flags"] & P_FLAGS.PF_W:
#    raise ValueError("Invalid ELF: contains writable segment")
# if elf_data:
#    raise ValueError("Invalid ELF: too many load segments")
# if not segment["p_flags"] & P_FLAGS.PF_X:
#    raise ValueError("Invalid ELF: no executable segment")
# if elf.entry < 0x3000:
#     raise RaspsimException("vaddr must be >= 0x3000")


@contextmanager
def rscompile(
    code: str,
    entry_label: str = "_start",
    lang: Literal["c", "assembler"] = "assembler",
    compiler: str = "cc",
    std: str = "c17",
    text_segment_addr: int = 0x4000,
    additional_flags: list[str] | None = None,
) -> Generator[IO[bytes], None, None]:
    """
    Compile assembly code to an ELF binary.

    Args:
        code: Assembly code to compile.
        entry_label: Entry point label.
        lang: Language of the code.
        compiler: Compiler to use.
        std: Standard to use.
        additional_flags: Additional flags to pass to the compiler.

    Yields:
        A file-like object containing the compiled ELF binary.
    """
    args = [
        compiler,
        f"-x{lang}",
        "-pipe",
        f"-std={std}",
        "-static",
        "-march=k8",
        "-mno-mmx",
        "-mno-80387",
        "-mtune=generic",  # prevent excessive loop unrolling
        "-ffreestanding",
        "-nostdlib",
        "-D_MM_MALLOC_H_INCLUDED",
        # Assembler arguments
        "-Wa,-march=k8+cmov+nommx",
        "-Wa,-mx86-used-note=no",
        "-Wa,--fatal-warnings",
        # Linker arguments
        "-Wl,-n",
        "-Wl,--fatal-warnings",  # fail if entry point doesn't exist
        "-Wl,--no-dynamic-linker",
        "-Wl,--build-id=none",
        "-Wl,-z,defs",
        "-Wl,-z,noexecstack",
        "-Wl,-z,norelro",
        "-Wl,-z,noseparate-code",
        f"-Wl,-Ttext={text_segment_addr:#x}",
        f"-Wl,-e,{entry_label}",
    ] + (additional_flags or [])
    with tempfile.NamedTemporaryFile("w+b") as f:
        subprocess.run(
            args + ["-o", f.name, "-"],
            input=code.encode(),
            check=True,
        )
        f.seek(0)
        yield f


@dataclass
class Segment:
    """
    A class representing a memory segment.
    Attributes:
        vaddr (int): The virtual address of the segment.
        prot (Prot): The protection attributes of the segment.
        size (int): The size of the segment.
        data (bytes | None): Optional data contained in the segment.
    Methods:
        __repr__(): Returns a string representation of the Segment instance.
    Note:
        Length of `data` shall not exceed the size of the segment.
        If size is greater than the length of the data, the rest of the segment denotes zeroed memory.
    """

    vaddr: int
    prot: Prot
    size: int
    data: bytes | None = None

    def __repr__(self):
        data_repr = f" data={self.data!r}" if self.data else ""
        return f"Segment(vaddr={self.vaddr:#x}, prot={self.prot}, size={self.size:#x}{data_repr})"


@dataclass
class ELF:
    """
    ELF binary data.

    Attributes:
        entry (int): The entry point of the ELF binary.
        stack (int | None): The stack pointer, if available.
        segments (list[Segment]): A list of segments loaded into memory.
    Methods:
        __repr__(): Returns a string representation of the ELF object.
    Note:
        Only contains the segments that are loaded into memory. Requires x86_64 executable ELF.
    """

    entry: int
    stack: int | None = None
    segments: list[Segment] = field(default_factory=list)

    def __repr__(self):
        return f"ELF(entry={self.entry:#x}, stack={self.stack:#x}, segments={self.segments})"


class InvalidELFException(Exception):
    """
    Exception raised when an invalid ELF file is encountered.
    """


def load_elf(
    elf: IO[bytes], gnu_stack_size: int = 0x1000, gnu_stack_addr: int = 0x7FE000000000
) -> ELF:
    """
    Load an ELF file and parse its segments.

    Args:
        elf (IO[bytes]): A byte stream representing the ELF file.
        gnu_stack_size (int, optional): The size of the stack when a GNU Stack Segment is found. Defaults to 0x1000.
        gnu_stack_addr (int, optional): The address of the segment when a GNU Stack Segment is found. Defaults to 0x7FE000000000.

    Returns:
        ELF: An ELF object containing the parsed segments and entry point.

    Raises:
        InvalidELFException: If the ELF file is not of type ET_EXEC or machine type EM_X86_64.
        InvalidELFException: If a segment's file size is greater than its memory size.
        ValueError: If the segment data length does not match the segment's file size.
        InvalidELFException: If an unhandled segment type is encountered.
    """
    # new stream from bytes
    elffile = ELFFile(elf)

    if elffile["e_type"] != "ET_EXEC":
        raise InvalidELFException(f"Invalid ELF type {elffile['e_type']}")

    if elffile["e_machine"] != "EM_X86_64":
        raise InvalidELFException(f"Invalid ELF machine {elffile['e_machine']}")

    loaded_elf = ELF(entry=elffile["e_entry"])

    for segment in elffile.iter_segments():
        logging.debug(
            "[.] ELF segment: %s V=%#x FSZ=%#x MSZ=%#x flags=%d",
            segment["p_type"],
            segment["p_vaddr"],
            segment["p_filesz"],
            segment["p_memsz"],
            segment["p_flags"],
        )
        if segment["p_type"] == "PT_LOAD":
            if segment["p_filesz"] > segment["p_memsz"]:
                raise InvalidELFException(
                    "Invalid ELF: corrupt binary. p_filesz > p_memsz"
                )

            # Try to read segment data
            segment_data = segment.data()
            if len(segment_data) != segment["p_filesz"]:
                raise ValueError("Invalid ELF: corrupt binary")

            logging.debug("[.] Read %d bytes of code/data", len(segment_data))
            loaded_elf.segments.append(
                Segment(
                    vaddr=segment["p_vaddr"],
                    size=segment["p_memsz"],
                    data=segment_data,
                    prot=getProtFromELFSegment(segment["p_flags"]),
                )
            )
        elif segment["p_type"] == "PT_GNU_STACK":
            loaded_elf.stack = gnu_stack_addr
            loaded_elf.segments.append(
                Segment(
                    vaddr=gnu_stack_addr - gnu_stack_size,
                    size=gnu_stack_size,
                    prot=Prot.RW,
                )
            )
        elif segment["p_type"] == "PT_NULL":
            pass
        else:
            raise InvalidELFException(
                f"Invalid ELF: unhandled segment {segment['p_type']}"
            )

    return loaded_elf


def elf_add_trampoline(elf: ELF):
    """
    Adds a trampoline to the ELF binary.
    
    The trampoline is a call instruction to the entry point followed by an terminate instruction.
    If reaching the terminator indicates that the program has gracefully returned from the call.
    
    Parameters:
        elf (ELF): The ELF object to which the trampoline will be added.
    Raises:
        RaspsimException: If the entry point is not page-aligned or if the entry 
                          point is less than 0x3000.
    
    Modifies:
        - Adjusts the entry point of the ELF object.
        - Appends a new segment containing the trampoline code to the ELF object's segments.
    """
    if elf.entry & 0xFFF:
        raise RaspsimException("code must be page-aligned")

    if elf.entry < 0x3000:
        raise RaspsimException("vaddr must be >= 0x3000 to use trampoline")

    trampoline_bias = 0x2000
    addr_start = max(0x1000, elf.entry - trampoline_bias)
    reloff = elf.entry - addr_start - 5
    trampoline = b"\xe8" + reloff.to_bytes(4, "little") + b"\xcd\x80"
    elf.segments.append(
        Segment(
            vaddr=addr_start, prot=Prot.RX, size=len(trampoline), data=trampoline
        )
    )
    elf.entry = addr_start

def asm_preamble(label: str = "_start") -> str:
    return cleandoc(f"""
                    .intel_syntax noprefix
                    .global {label}
                    .section .text
                    {label}:
                    """)

def asm_stop_sim() -> str:
    return "int 0x80\n"

def populate_from_elf(sim: Raspsim, elf: ELF, abi: Literal["sysv"] = "sysv"):
    """
    Populate the simulator with the ELF binary data.

    Args:
        sim: Raspsim instance.
        elf: ELF binary data.
        abi: ABI to use. Currently only supports sysv.
    """
    del abi  # unused

    sim.rip = elf.entry

    if elf.stack is None:
        raise RaspsimException("stack must be set")
    sim.rsp = elf.stack

    for segment in elf.segments:
        sim.memmap(
            segment.vaddr, segment.prot, length=segment.size, data=segment.data or b""
        )



import sys
from .utils import rscompile, populate_from_elf, load_elf
from .core import Raspsim

if __name__ == "__main__":
    # check if arguments given
    if len(sys.argv) > 1:
        print("Usage: python -m pyraspsim < <file>\n\nRunning as module reads from stdin and does not take arguments.", file=sys.stderr)

    program = "\n".join(sys.stdin.readlines())

    with rscompile(program) as f:
        elf = load_elf(f)
        
    # Create a new Raspsim instance
    sim = Raspsim()

    populate_from_elf(sim, elf)

    # Run the simulation
    sim.run()

    # Print the simulation state
    print(sim)

    


try:
    from .core import *
except ImportError as e:
    raise ImportError("""Failed to import core module. This is most likely due to an unsupoorted python version, plattform or architecture.
    Supported versions are 3.7 to 3.12 on linux x86_64""") from e


"""
python binding for raspsim, a cycle-accurate x86 simulator based on PTLsim
"""
from __future__ import annotations
import typing
__all__ = ['Address', 'BoundsException', 'BreakpointException', 'CoprocOverrunException', 'DebugException', 'DivideException', 'DoubleFaultException', 'FPUException', 'FPUNotAvailException', 'GPFaultException', 'InvalidOpcodeException', 'InvalidTSSException', 'MachineCheckException', 'NMIException', 'OverflowException', 'PageFaultException', 'Prot', 'Raspsim', 'RaspsimException', 'SSEException', 'SegNotPresentException', 'SpuriousIntException', 'StackFaultException', 'UnalignedException', 'getProtFromELFSegment']
class Address:
    @staticmethod
    def _pybind11_conduit_v1_(*args, **kwargs):
        ...
    def __getitem__(self, offset: int) -> Address:
        ...
    @typing.overload
    def __setitem__(self, offset: int, value: bytes) -> None:
        ...
    @typing.overload
    def __setitem__(self, offset: int, value: int) -> None:
        ...
    def read(self, size: int = 1) -> bytes:
        """
        Read data from the address
        """
    def write(self, value: bytes) -> None:
        """
        Write data to the address
        """
class BoundsException(Exception):
    pass
class BreakpointException(Exception):
    pass
class CoprocOverrunException(Exception):
    pass
class DebugException(Exception):
    pass
class DivideException(Exception):
    pass
class DoubleFaultException(Exception):
    pass
class FPUException(Exception):
    pass
class FPUNotAvailException(Exception):
    pass
class GPFaultException(Exception):
    pass
class InvalidOpcodeException(Exception):
    pass
class InvalidTSSException(Exception):
    pass
class MachineCheckException(Exception):
    pass
class NMIException(Exception):
    pass
class OverflowException(Exception):
    pass
class PageFaultException(Exception):
    pass
class Prot:
    """
    Members:
    
      READ
    
      WRITE
    
      EXEC
    
      NONE
    
      RW
    
      RX
    
      RWX
    """
    EXEC: typing.ClassVar[Prot]  # value = <Prot.EXEC: 4>
    NONE: typing.ClassVar[Prot]  # value = <Prot.NONE: 0>
    READ: typing.ClassVar[Prot]  # value = <Prot.READ: 1>
    RW: typing.ClassVar[Prot]  # value = <Prot.RW: 3>
    RWX: typing.ClassVar[Prot]  # value = <Prot.RWX: 7>
    RX: typing.ClassVar[Prot]  # value = <Prot.RX: 5>
    WRITE: typing.ClassVar[Prot]  # value = <Prot.WRITE: 2>
    __members__: typing.ClassVar[dict[str, Prot]]  # value = {'READ': <Prot.READ: 1>, 'WRITE': <Prot.WRITE: 2>, 'EXEC': <Prot.EXEC: 4>, 'NONE': <Prot.NONE: 0>, 'RW': <Prot.RW: 3>, 'RX': <Prot.RX: 5>, 'RWX': <Prot.RWX: 7>}
    @staticmethod
    def _pybind11_conduit_v1_(*args, **kwargs):
        ...
    def __eq__(self, other: typing.Any) -> bool:
        ...
    def __getstate__(self) -> int:
        ...
    def __hash__(self) -> int:
        ...
    def __index__(self) -> int:
        ...
    def __init__(self, value: int) -> None:
        ...
    def __int__(self) -> int:
        ...
    def __ne__(self, other: typing.Any) -> bool:
        ...
    def __repr__(self) -> str:
        ...
    def __setstate__(self, state: int) -> None:
        ...
    def __str__(self) -> str:
        ...
    @property
    def name(self) -> str:
        ...
    @property
    def value(self) -> int:
        ...
class Raspsim:
    r10: int
    r11: int
    r12: int
    r13: int
    r14: int
    r15: int
    r8: int
    r9: int
    rax: int
    rbp: int
    rbx: int
    rcx: int
    rdi: int
    rdx: int
    rip: int
    rsi: int
    rsp: int
    @staticmethod
    def _pybind11_conduit_v1_(*args, **kwargs):
        ...
    def __init__(self, logfile: str = '/dev/null') -> None:
        """
        Create a new Raspsim instance
        """
    def __str__(self) -> str:
        """
        Get the string representation of the current state of the simulator
        """
    def disableSSE(self) -> None:
        """
        Disable SSE
        """
    def disableX87(self) -> None:
        """
        Disable X87
        """
    def enablePerfectCache(self) -> None:
        """
        Enable perfect cache
        """
    def enableStaticBranchPrediction(self) -> None:
        """
        Enable static branch prediction
        """
    def getMappedPage(self, arg0: int) -> Address:
        """
        Get the mapped page
        """
    def mapPage(self, start: int, prot: Prot) -> Address:
        """
        Map a page to the virtual address space of the simulator
        """
    def memmap(self, start: int, prot: Prot, length: int = 0, data: bytes = b'') -> Address:
        """
        Map a range of memory to the virtual address space of the simulator.
        
        Maps data from `data` into memory and fills the rest with zeros if `length` is greater than the size of `data`. If `length` is 0, the size of `data` will be used as length.
        """
    def run(self) -> None:
        """
        Run the simulator
        """
    @property
    def cycles(self) -> int:
        """
        Get the number of cycles
        """
    @property
    def instructions(self) -> int:
        """
        Get the number of instructions
        """
class RaspsimException(Exception):
    pass
class SSEException(Exception):
    pass
class SegNotPresentException(Exception):
    pass
class SpuriousIntException(Exception):
    pass
class StackFaultException(Exception):
    pass
class UnalignedException(Exception):
    pass
def getProtFromELFSegment(flags: int) -> Prot:
    """
    Get the protection as Raspsim Prot from ELF segment flags
    """



def iNty(n: int):
    def wrapper(cls):
        class iNty(cls):
            def __init__(self, value: int | None = None):
                if value is None:
                    value = 0
                
                self.value = value % (2 ** n)

            def __repr__(self):
                negrep = ""
                if self.value >> (n - 1):
                    negrep = ", -" + str(2 ** n - self.value)
                    
                return f"i{n}({self.value}, {self.value:#x}{negrep})"

            def __add__(self, other: cls):
                return i64(self.value + other.value)

            def __sub__(self, other: cls):
                return i64(self.value - other.value)

            def __mul__(self, other: cls):
                return iNty(self.value * other.value)

            def __truediv__(self, other: cls):
                return iNty(self.value // other.value)

            def __floordiv__(self, other: cls):
                return iNty(self.value // other.value)

            def __mod__(self, other: cls):
                return iNty(self.value % other.value)

            def __pow__(self, other: cls):
                return iNty(self.value ** other.value)

            def __lshift__(self, other: cls):
                return iNty(self.value << other.value)

            def __rshift__(self, other: cls):
                return iNty(self.value >> other.value)

            def __and__(self, other: cls):
                return iNty(self.value & other.value)

            def __or__(self, other: cls):
                return iNty(self.value | other.value)

            def __xor__(self, other: cls):
                return iNty(self.value ^ other.value)

            def __neg__(self):
                return iNty(-self.value)

            def __pos__(self):
                return iNty(+self.value)

            def __invert__(self):
                return iNty(~self.value)

            def __eq__(self, other: int | cls):
                if isinstance(other, int):
                    return self.value == other
                return self.value == other.value
            
            def __ne__(self, other: int | cls):
                if isinstance(other, int):
                    return self.value != other
                return self.value != other.value
            
            def __lt__(self, other: int | cls):
                if isinstance(other, int):
                    return self.value <  other
                return self.value < other.value
            
            def __le__(self, other: int | cls):
                if isinstance(other, int):
                    return self.value <= other
                return self.value <= other.value
            
            def __gt__(self, other: int | cls):
                if isinstance(other, int):
                    return self.value >  other
                return self.value > other.value
            
            def __ge__(self, other: int | cls):
                if isinstance(other, int):
                    return self.value >= other
                return self.value >= other.value
            
            def __bool__(self):
                return bool(self.value)

            def __int__(self):
                return self.value
        
        iNty.__name__ = cls.__name__
        return iNty
    return wrapper
    

@iNty(8)
class i8:
    """
    A class representing an 8-bit integer.
    """

@iNty(16)
class i16:
    """
    A class representing a 16-bit integer.
    """

@iNty(32)
class i32:
    """
    A class representing a 32-bit integer.
    """

@iNty(64)
class i64:
    """
    A class representing a 64-bit integer.
    """
